#include "downloader.h"

#include <iostream>

#include <QEventLoop>
#include <QJsonArray>
#include <QJsonDocument>


// A simple object that handles HTTP(S) GET/POST requests, sending the fetched via a
// signal (downloadFinished()) when complete. Only a single object at once.
Downloader::~Downloader()
{
    delete fetchedData;
    netReply->deleteLater();
}


// Start a download, with the given request (containing the URL and any desired headers)
// and optional postData for POST. If it is nullptr, then the request is assumed to be GET.
void
Downloader::startDownload(QNetworkRequest* request, QByteArray* postData)
{
    fetchedData = new QByteArray();

    emit Downloader::downloadProgress(false, "Připojování k serveru…");
    request->setRawHeader("User-Agent", "TSPbrousek");

    if (postData != nullptr)
	netReply = netManager.post(*request, *postData);
    else
	netReply = netManager.get(*request);

    connect(netReply, &QIODevice::readyRead,
	    this, &Downloader::httpRead);
    connect(netReply, &QNetworkReply::downloadProgress,
	    this, &Downloader::httpProgress);
    connect(netReply, &QNetworkReply::finished,
	    this, &Downloader::httpFinished);
    connect(netReply, &QNetworkReply::errorOccurred,
	    this, &Downloader::httpError);

    // Block current thread until the download has finished.
    QEventLoop downloadLoop;
    connect(netReply, &QNetworkReply::finished,
	    &downloadLoop, &QEventLoop::quit);
    downloadLoop.exec();
}


// Save any data from the HTTP request.
// Handles signal QIODevice::readyRead (netReply).
void
Downloader::httpRead()
{
    fetchedData->append(netReply->readAll());
}


// Make sure to broadcast that the connection has been established.
// Handles signal QNetworkReply::downloadProgress (netReply).
void
Downloader::httpProgress(qint64 received, qint64 total)
{
    emit Downloader::downloadProgress(true, "Získávání odpovědi…");
}


// Broadcast any connection/HTTP errors that might be encoutered.
// Handles signal QNetworkReply::errorOccured (netReply).
void
Downloader::httpError(QNetworkReply::NetworkError code)
{
    QString errorMsg("Získávání údajů od serveru se nezdařilo. Jste připojeni k internetu? Je zvolená služba dostupná?\r\n");
    errorMsg.append(netReply->errorString());
    emit Downloader::downloadError(errorMsg);
}


// When the download has finished, make sure to broadcast all accumulated data.
// Handles signal QNetworkReply::finished (netReply).
void
Downloader::httpFinished()
{
    emit Downloader::downloadFinished(fetchedData);
}
