#include "netrouter.h"

#include <iostream>

#include <QEventLoop>
#include <QJsonArray>
#include <QJsonDocument>

#include "downloader.h"


// A class for all Routers that use HTTP APIs, adding a generic interface for
// downloading and parsing the relevant data.
NetRouter::NetRouter()
{
    downloader = new Downloader();
}


NetRouter::~NetRouter()
{
    delete downloader;
}


// The template for downloading/parsing to be used by all NetRouters.
// The invoked methods `parseDistancesFromResponse` and `makeRoutingRequest`
// are how any engine subclasses actually implement their distance calculations.

void
NetRouter::calculateDistances(vector_2d_double points)
{
    currentPoints = points;

    connect(downloader, &Downloader::downloadProgress,
	    this, &NetRouter::downloadProgress);
    connect(downloader, &Downloader::downloadError,
	    this, &NetRouter::downloadError);
    connect(downloader, &Downloader::downloadFinished,
	    this, &NetRouter::parseDistancesFromResponse);

    downloader->startDownload(makeRoutingRequest(points), makePostData(points));
}


// This is the function responsible for parsing received data; if overloaded by an
// engine subclass, keep in mind that Router::calculationError or Router::distancesCalculated
// must be emitted.
// For now, both NetRouters (OSRM/FOSSGIS and OpenRouterService) return JSON of the same
// format, so we need only define the parsing here.
// Handles signal Downloader::downloadFinished (downloader).
void
NetRouter::parseDistancesFromResponse(QByteArray* data)
{
    emit Router::calculationProgress(50, 0, "Načítání odpovědi ve formátu JSON…");

    QJsonParseError* jsonError = nullptr;
    QJsonDocument json = QJsonDocument::fromJson(*data, jsonError);

    emit Router::calculationProgress(75, 0, "Extrakce vzdáleností z formátu JSON…");

    QString errorMsg("Obdržené údaje nejsou ve formátu JSON.\r\nJste si jisti, že daná adresa skutečně odkazuje na zvolenou službu?\r\n");
    if (json.isNull()) {
	if (jsonError != nullptr)
	    errorMsg.append(jsonError->errorString());
	emit Router::calculationError(errorMsg);
	return;
    }

    if (json["error"].isString()) {
	errorMsg = generic_api_error_to_message(json["error"].toString());
	emit Router::calculationError(errorMsg);
	return;
    }

    errorMsg = "Obdržené údaje nejsou v tom správném formátu.\r\nJste si jisti, že daná adresa skutečně odkazuje na zvolenou službu?\r\n";

    if (json["distances"].isNull() || json["distances"].isArray() == false) {
	emit Router::calculationError(errorMsg);
	return;
    }

    vector_2d_int distances;
    QJsonArray distanceArray = json["distances"].toArray();

    int pointc = currentPoints.size();
    if (distanceArray.count() != pointc) {
	emit Router::calculationError(errorMsg);
	return;
    }

    for (int i = 0; i < pointc; i++) {
	std::vector<int> distanceRow;
	QJsonArray distanceRowArray = distanceArray[i].toArray();
	if (distanceRowArray.count() != pointc) {
	    emit Router::calculationError(errorMsg);
	    return;
	}

	for (int j = 0; j < pointc; j++) {
	    QJsonValue value = distanceRowArray[j];
	    if (value.isDouble() == false) {
		emit Router::calculationError(errorMsg);
		return;
	    }
	    distanceRow.push_back(value.toDouble());
	}
	distances.push_back(distanceRow);
    }

    emit Router::distancesCalculated(currentPoints, distances);
}


// Report any downloading errors.
// Handles signal Downloader::downloadError (downloader).
void
NetRouter::downloadError(QString message)
{
    emit Router::calculationError(message);
}


// Report download progress (connecting/connected).
// Handles signal Downloader::downloadProgress (downloader).
void
NetRouter::downloadProgress(bool connected, QString message)
{
    emit Router::calculationProgress(connected ? 20 : 0, 0, message);
}
