#include "openrouteservice.h"

#include <iostream>

#include <QEventLoop>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>

#include "addressselectionwindow.h"


// Distance-calculation using OpenRouteService's HTTP API.
OpenRouteServiceRouter::OpenRouteServiceRouter(OpenRouteServiceMapData* map_data)
{
    mapData = map_data;
}


// Return a “matrix” routing URL as per the OpenRouteService API, with the given
// points as parameters.
QNetworkRequest*
OpenRouteServiceRouter::makeRoutingRequest(vector_2d_double points)
{
    QString url = mapData->serverUrl;
    url.append("/v2/matrix/");

    if (mapData->byCar)
	url.append("driving-car");
    else
	url.append("foot-walking");

    QNetworkRequest* request = new QNetworkRequest(url);
    request->setHeader(QNetworkRequest::ContentTypeHeader, "application/json");
    request->setRawHeader("Authorization", mapData->apiKey.toUtf8());

    return request;
}


// Create the appropriate POST data for an ORS “matrix” request.
QByteArray*
OpenRouteServiceRouter::makePostData(vector_2d_double points)
{
    QString jsonPayload("{\"metrics\":[\"distance\"],\"locations\":[");
    for (int i = 0; i < points.size(); i++) {
	if (i > 0)
	    jsonPayload.append(",");

	QString parameter =
	    QString("[%1,%2]").arg(points[i][1], 0, 'g', 9).arg(points[i][0], 0, 'g', 9);
	jsonPayload.append(parameter);
    }
    jsonPayload.append("]};");

    QByteArray* data = new QByteArray();
    *data = jsonPayload.toUtf8();
    return data;
}



// “Addressing” (geocoding) using OpenRouteService's HTTP API.
OpenRouteServiceAddresser::OpenRouteServiceAddresser(OpenRouteServiceMapData* map_data)
{
    mapData = map_data;
}


// Parse the server's response (JSON) into useful coordinates.
// Emits Addresser::locationError and Addresser::locationFound.
void
OpenRouteServiceAddresser::parseCoordinatesFromResponse(QByteArray* data)
{
    QJsonParseError* jsonError = nullptr;
    QJsonDocument json = QJsonDocument::fromJson(*data, jsonError);

    QString errorMsg("Obdržené údaje nejsou ve formátu JSON.\r\nJste si jisti, "
		     "že daná adresa skutečně odkazuje na službu OpenRouteService?\r\n");
    if (json.isNull()) {
	if (jsonError != nullptr)
	    errorMsg.append(jsonError->errorString());
	emit Addresser::locationError(errorMsg);
	return;
    }

    if (json["error"].isString()) {
	QString errorMsg = generic_api_error_to_message(json["error"].toString());
	emit Addresser::locationError(errorMsg);
	return;
    }

    errorMsg = "Obdržené údaje nejsou v tom správném formátu.\r\nJste si jisti, "
	"že daná adresa skutečně odkazuje na službu OpenRouteService?\r\n";
    if (json["features"].isArray() == false) {
	emit Addresser::locationError(errorMsg);
	return;
    }

    QJsonValue features = json["features"].toArray();
    if (features.toArray().empty()) {
	emit Addresser::locationError("Pro danou adresu nebyla nalezena žádná lokace.");
	return;
    }

    // Collect all geocode results…
    vector_2d_double points;
    QStringList pointNames;
    for (int i = 0; i < features.toArray().count(); i++) {
	QJsonArray coords = features[i]["geometry"]["coordinates"].toArray();
	std::vector<double> point = { coords[1].toDouble(), coords[0].toDouble() };
	points.push_back(point);

	QJsonObject properties = features[i]["properties"].toObject();
	if (properties["label"].isString())
	    pointNames.append(properties["label"].toString());
	else if (properties["name"].isString())
	    pointNames.append(properties["name"].toString());
	else
	    pointNames.append("- Bez názvu -");
    }

    if (points.size() == 1) {
	emit Addresser::locationFound(points[0], pointNames[0]);
	return;
    }

    // If there're more than one possible points, we need to ask the user.
    AddressSelectionWindow* selector = new AddressSelectionWindow(nullptr, points, pointNames);
    connect(selector, &AddressSelectionWindow::addressSelected,
	    this, &OpenRouteServiceAddresser::addressSelected);
    connect(selector, &QDialog::rejected,
	    this, &OpenRouteServiceAddresser::noAddressSelected);
    selector->show();
}


// Create the network request (URL and headers) appropriate for geocoding with ORS.
QNetworkRequest*
OpenRouteServiceAddresser::makeLocatingRequest(QString address)
{
    QString url = mapData->serverUrl;
    url.append("/geocode/search?api_key=");
    url.append(mapData->apiKey);
    url.append("&text=");
    url.append(address);

    QNetworkRequest* request = new QNetworkRequest(url);
    request->setHeader(QNetworkRequest::ContentTypeHeader, "application/json");
    return request;
}


// Sends the given point and point-name in the locationFound signal; this is called
// by an AddressSelectionWindow when there are more than one possible points according
// and address.
// Handles signal AddressSelectWindow::addressSelected.
void
OpenRouteServiceAddresser::addressSelected(std::vector<double> point, QString name)
{
    emit Addresser::locationFound(point, name);
}


// Even if the user selects no result, we've got to inform any listeners.
// Handles signal QDialog::rejected (AddressSelectWindow).
void
OpenRouteServiceAddresser::noAddressSelected()
{
    emit Addresser::locationCancelled();
}
